import ipywidgets as widgets
from ipywidgets import VBox, Layout, HBox
from IPython.core.display import clear_output
from src.replifactory import MorbidostatCulture
from src.replifactory import TurbidostatCulture
from src.replifactory import EnduranceStressCulture
from src.replifactory.culture.blank import BlankCulture
from src.replifactory import BatchCulture
from src.replifactory import PatientCulture
from src.replifactory import Experiment

import os
import glob


class ProgramViewerTab:
    title = "Program Viewer"

    def __init__(self, main_gui):
        self.main_gui = main_gui
        self.title = "Experiment config"

        config_paths = glob.glob("../**/device_config.yaml", recursive=True)
        self.config_paths = [os.path.relpath(p) for p in config_paths]
        self.experiment_directories = [os.path.relpath(os.path.join(p, "..")) for p in self.config_paths]
        self.load_exp = widgets.Dropdown(options=self.experiment_directories, description="Load experiment",
                                         style={"description_width": "initial"}, index=None)
        self.new_exp_folder = widgets.Text(description="new experiment directory:",
                                           style={"description_width": "initial"})
        make_new_exp = widgets.Button(description="new experiment")
        make_new_exp.on_click(self.handle_new_experiment)
        self.load_exp.observe(self.handle_experiment_folder_choice, names="value")
        self.selector = VBox([self.load_exp, HBox([self.new_exp_folder, make_new_exp])])
        self.vials = widgets.Label("No experiment selected")
        self.widget = VBox([self.selector, self.vials])

    def handle_new_experiment(self, b):
        if self.main_gui.device is None:
            with self.main_gui.status_bar.output:
                print("PLEASE CONNECT DEVICE TO CREATE NEW EXPERIMENT!")
        else:
            with self.main_gui.status_bar.output:
                clear_output()
                self.main_gui.experiment = Experiment(self.new_exp_folder.value)
                self.update()

    def handle_experiment_folder_choice(self, change):
        with self.main_gui.status_bar.output:
            clear_output()
            self.main_gui.experiment = Experiment(change.new)
        self.update()

    def make_vial_widgets(self):
        # if self.main_gui.experiment is not None:
        #     if self.main_gui.experiment.device is not None:
        title = widgets.HTML('<b>Selected experiment directory:   %s:</b>' % self.main_gui.experiment.directory,
                             layout=Layout(width="400px"))
        self.vials = VBox([title] + [VialConfigWidget(program_viewer_tab=self, vial_number=v).widget for v in range(1, 8)])
    def update(self):
        self.make_vial_widgets()
        self.widget = VBox([self.selector, self.vials])
        self.main_gui.update()

    # def make_vial_widget(self, vial_number):
    #     vial = self.main_gui.experiment.device.cultures[vial_number]
    #     title = widgets.HTML('<b>Vial %d:</b>' % vial_number, layout=Layout(width="40px", height="40px"))
    #     culture_options = [BlankCulture, TurbidostatCulture, MorbidostatCulture, EnduranceStressCulture]
    #     culture_options_names = [str(klass)[8:-2] for klass in culture_options]
    #     culture_options += [type(None)]
    #     culture_options_names += ["None"]
    #
    #     vial_culture_index = culture_options.index(type(vial))
    #     algorithm = widgets.Dropdown(options=list(zip(culture_options_names, culture_options)),
    #                                  index=vial_culture_index)
    #
    #     def handle_vial_class_change(change):
    #         if change.new is type(None):
    #             self.main_gui.experiment.device.cultures[vial_number] = None
    #         else:
    #             self.main_gui.experiment.device.cultures[vial_number] = \
    #                 change.new(directory=self.main_gui.experiment.directory, vial_number=vial_number)
    #         self.widget = self.make_vial_widgets()
    #         self.main_gui.update()
    #
    #     algorithm.observe(handle_vial_class_change, names="value")
    #
    #     if vial is not None:
    #         description_style = {}
    #         description_widgets = [
    #             widgets.Text(value=vial.name, description="name", style=description_style, continuous_update=True),
    #             widgets.Textarea(value=vial.description, description="description",
    #                              style=description_style, continuous_update=True, layout=Layout(height="80px"))]
    #         for w in description_widgets:
    #             w.observe(vial.handle_value_change, names="value")
    #         description_box = widgets.VBox(description_widgets)
    #         left_panel = VBox([title, algorithm, description_box])
    #         return HBox([left_panel, vial.widget])
    #     else:
    #         return VBox([title, algorithm])


class VialConfigWidget:
    def __init__(self, program_viewer_tab, vial_number):
        self.program_viewer_tab = program_viewer_tab
        self.main_gui = program_viewer_tab.main_gui
        self.vial_number = vial_number
        self.culture = self.main_gui.experiment.device.cultures[vial_number]
        title = widgets.HTML('<b>Vial %d:</b>' % vial_number, layout=Layout(width="40px", height="40px"))
        self.inoculate = widgets.Button(description="inoculate")
        self.inoculate.on_click(self.handle_inoculate_button)
        if type(self.culture) not in [BlankCulture, type(None)]:
            if self.culture._inoculation_time:
                self.inoculate.disabled = True
                self.inoculate.description = "inoculated"
        self.blank = widgets.Button(description="OD BLANK")
        self.blank.on_click(self.handle_blank_button)

        culture_options = [BlankCulture, BatchCulture, TurbidostatCulture, MorbidostatCulture, EnduranceStressCulture,
                           PatientCulture]
        culture_options_names = [str(klass)[8:-2] for klass in culture_options]
        culture_options += [type(None)]
        culture_options_names += ["None"]

        culture_options_index = culture_options.index(type(self.culture))
        algorithm = widgets.Dropdown(options=list(zip(culture_options_names, culture_options)),
                                     index=culture_options_index)

        algorithm.observe(self.handle_vial_class_change, names="value")

        if self.culture is not None:
            description_style = {}
            description_widgets = [
                widgets.Text(value=self.culture.name, description="name", style=description_style, continuous_update=True),
                widgets.Textarea(value=self.culture.description, description="description",
                                 style=description_style, continuous_update=True, layout=Layout(height="80px"))]
            for w in description_widgets:
                w.observe(self.handle_vial_parameter_change, names="value")
            description_box = widgets.VBox(description_widgets, layout=Layout(height="150px"))
            left_panel = VBox([HBox([title, self.blank, self.inoculate]), algorithm, description_box])
            self.widget = HBox([left_panel, self.culture.widget], layout=Layout(border='solid'))
        else:
            self.widget = VBox([title, algorithm], layout=Layout(border='solid'))

    def handle_blank_button(self, button):
        self.culture.write_blank_od()
        self.program_viewer_tab.update()

    def handle_inoculate_button(self, button):
        self.culture.inoculate()
        button.disabled = True
        button.description = "inoculated"
        self.culture.save()

    def handle_vial_parameter_change(self, change):
        parameter_name = change.owner.description
        self.culture.__dict__[parameter_name] = change.new
        self.culture.save()

    def handle_vial_class_change(self, change):
        if change.new is type(None):
            self.main_gui.experiment.device.cultures[self.vial_number] = None
        else:
            self.main_gui.experiment.device.cultures[self.vial_number] = \
                change.new(directory=self.main_gui.experiment.directory, vial_number=self.vial_number)
        self.widget = self.__init__(program_viewer_tab=self.program_viewer_tab, vial_number=self.vial_number)
        self.program_viewer_tab.update()



import os
from src.replifactory import adaptive_window_doubling_time
from src.replifactory import read_csv_tail
import time
import numpy as np
import matplotlib.pyplot as plt


def plot_culture(culture, last_hours=24, plot_growth_rate=False):
    od_filepath = os.path.join(culture.directory, "od.csv")
    df = read_csv_tail(filepath=od_filepath, lines=last_hours * 60)
    t = df.index.values
    od = df.values.ravel()
    t_min = t[-1] - last_hours * 3600
    odw = od[t > t_min]
    tw = t[t > t_min]
    drug_concentration_file = os.path.join(culture.directory, "medium2_concentration.csv")
    tdose = None
    dosevalues = None
    if os.path.exists(drug_concentration_file):
        dosedf = read_csv_tail(drug_concentration_file, lines=50 + last_hours * 10)
        tdose = dosedf.index.values
        dosevalues = dosedf.values.ravel()[tdose > t_min]
        tdose = tdose[tdose > t_min]

    # fig = plot_gr(tw, odw, tdose)

    # plot growth rate
    od_values = odw
    time_values = tw
    dilution_timepoints = tdose
    od = np.array(od_values)
    t = np.array(time_values)

    fig, ax = plt.subplots(figsize=[16, 8], dpi=100)
    # i = 0
    # lines = []
    ax.plot(t / 3600, od, "k.", label="Optical Density")
    # ax.set_ylim(-0.05, 1.6)
    od[od <= 0] = 1e-6
    ax.set_ylabel("Optical Density")
    ax.set_xlabel("Time [hours]")

    # Growth rate
    if plot_growth_rate:
        ax2 = ax.twinx()
        #
        td_timepoints, td, tderr = adaptive_window_doubling_time(t, od, dilution_timepoints=dilution_timepoints)

        markers, caps, bars = ax2.errorbar(td_timepoints / 3600, td, tderr,
                                           alpha=0.5, label="doubling time")
        [bar.set_alpha(0.1) for bar in bars]
        try:
            td = np.array(td)
            tderr = np.array(tderr)
            tdmax = np.nanmax(td[tderr < 0.05])
            tdmin = np.nanmin(td[tderr < 0.05])
            ax2.set_ylim(tdmin * 0.5, tdmax * 1.2)
        except:
            pass
        ax2.grid(color="xkcd:light blue", linestyle=":", alpha=0.8)
        ax2.yaxis.set_tick_params(color="xkcd:cerulean")
        ax2.set_ylabel("Doubling time [hours]", color="xkcd:cerulean")

    if tdose is not None:
        ax3 = fig.axes[0].twinx()
        ax3.step(tdose / 3600, dosevalues, "r^-", where="post", label="dose")
        ax3.spines["left"].set_position(("axes", -0.1))
        ax3.yaxis.set_label_position("left")
        ax3.yaxis.set_ticks_position("left")
        ax3.yaxis.set_tick_params(color="r")
        ax3.set_ylabel("Dose [mM]", color="r")
        ax3.grid(color="r", linestyle=":", alpha=0.5)

    handles, labels = [], []
    for axis in fig.axes:
        handle, label = axis.get_legend_handles_labels()
        axis.legend([])
        handles += handle
        labels += label
    ax.legend(handles, labels, loc=2)

    xticks = ax.get_xticks()
    for axis in fig.axes:
        axis.set_xticks(xticks)

    tmin = xticks[0]
    xtick_labels = [round(t - tmin, 2) for t in xticks]
    # fig.axes[1].set_ylim(0.1, 10)
    ax.set_xticklabels(xtick_labels)
    # fig.axes[1].set_yticks([])

    ax.set_yscale("log")
    od_ticks = [0.001, 0.01, 0.1, 0.15, 0.2, 0.3, 0.4, 0.5, 1]
    ax.set_yticks(od_ticks)
    ax.set_yticklabels(od_ticks)
    ax.set_ylim(0.0008, 1.6)
    ax.grid(color="k", linestyle="--", alpha=0.3)
    ax.set_xlabel("Time [hours from %s]" % time.ctime(tmin * 3600))

    plot_title = "%s %s" % (culture.name, culture.directory)
    fig.axes[0].set_title(plot_title)
    return fig


def plot_temperature(culture, fig, last_hours):
    df = read_csv_tail(os.path.join(culture.device.directory, "temperature.csv"), last_hours * 60)
    ax4 = fig.axes[0].twinx()
    ax4.plot(df.index.values/3600, df.temperature_vials, ".", label="temperature")
    ax4.spines["right"].set_position(("axes", 0.1))
    ax4.yaxis.set_label_position("right")
    ax4.yaxis.set_ticks_position("right")
    ax4.legend()



from .device.base_device import BaseDevice
from .device.morbidostat_device import MorbidostatDevice
from .culture import TurbidostatCulture, MorbidostatCulture
from .experiment import Experiment
from .GUI.main_gui import MainGui as Gui
from .GUI.notifier import Notifier
from . import util



import time
import pyftdi.i2c
from pyftdi.spi import SpiController
import threading
import os
import yaml
from matplotlib import pyplot as plt

from src.replifactory import Photodiodes
from src.replifactory import Lasers
from src.replifactory import OdSensor, measure_od_all
from src.replifactory.device.pwm import PwmController
from src.replifactory.device.stirrers import Stirrers
from src.replifactory import Pump, calibrate_pump
from src.replifactory.device.thermometers import Thermometers
from src.replifactory import Valves
from src.replifactory import make_dilution
from src.replifactory import CultureDict
from src.replifactory.loading import save_object, load_config, load_object
from src.replifactory import QueueWorker


class BaseDevice:
    PORT_ADC = 0x68  # MCP3421A0  1101 000
    # PORT_ADC = 0x69  # MCP3421A1  1101 001
    # PORT_ADC = 0x6a  # MCP3421A2  1101 010
    # PORT_ADC = 0x6b  # MCP3421A3  1101 011
    PORT_GPIO_MULTIPLEXER_LASERS = 0x20  # PCA 9555
    PORT_GPIO_MULTIPLEXER_ADC = 0x21  # PCA 9555
    PORT_GPIO_MULTIPLEXER_STIRRERS = 0x25  # PCA 9555
    PORT_THERMOMETER_VIALS = 0x48  # ADT 75
    PORT_THERMOMETER_BOARD = 0x4C  # ADT 75
    PORT_PWM = 0x70  # PCA9685

    def __init__(self, ftdi_address='ftdi://ftdi:2232h', connect=False):
        self.ftdi_address = ftdi_address
        self.directory = None
        self.cultures = CultureDict(self)
        self.cultures[1] = None
        self.cultures[2] = None
        self.cultures[3] = None
        self.cultures[4] = None
        self.cultures[5] = None
        self.cultures[6] = None
        self.cultures[7] = None

        self.dilution_worker = None
        self.od_worker = None

        self.hard_stop_trigger = False
        self.soft_stop_trigger = False

        self.i2c = None
        self.spi = None

        self.calibration_mv_to_od = {1: {}, 2: {}, 3: {}, 4: {}, 5: {}, 6: {}, 7: {}}
        self.calibration_fan_speed_to_duty_cycle = {1: {1: None, 2: None, 3: None},
                                                    2: {1: None, 2: None, 3: None},
                                                    3: {1: None, 2: None, 3: None},
                                                    4: {1: None, 2: None, 3: None},
                                                    5: {1: None, 2: None, 3: None},
                                                    6: {1: None, 2: None, 3: None},
                                                    7: {1: None, 2: None, 3: None}}
        self.calibration_pump_rotations_to_ml = {1: {},
                                                 2: {},
                                                 3: {},
                                                 4: {}}

        self.delay_tube_wash_hrs = 24
        self.pump1_flush_volume = 1
        self.pump2_flush_volume = 0.1
        self.pump3_flush_volume = 0

        self.locks_vials = {v: threading.Lock() for v in range(1, 8)}
        self.lock_pumps = threading.Lock()
        self.file_lock = threading.Lock()

        # self.pump_calibrations_rotations_to_ml = {1: {}, 2: {}, 3: {}, 4: {}}
        self.pump_stock_concentrations = {1: None, 2: None, 3: None, 4: None}
        self.pump_stock_volumes = {1: None, 2: None, 3: None, 4: None}

        self.od_values = {v: None for v in range(1, 8)}
        self.pwm_controller = PwmController(device=self, frequency=50)
        self.valves = Valves(device=self)
        self.stirrers = Stirrers(device=self)
        self.photodiodes = Photodiodes(device=self)
        self.lasers = Lasers(device=self)
        self.od_sensors = {v: OdSensor(device=self, vial_number=v) for v in range(1, 8)}
        self.pump1 = Pump(device=self, cs=0)
        self.pump2 = Pump(device=self, cs=1)
        self.pump3 = Pump(device=self, cs=2)
        self.pump4 = Pump(device=self, cs=3)
        self.thermometers = Thermometers(device=self)
        # if self.ftdi_address is not None:
        #     self.connect()
        if self.directory is not None:
            try:
                self.load()
            except FileNotFoundError:
                self.save()
        if connect:
            self.connect()

    def connect(self, ftdi_address=None, fit_calibration=False):
        if ftdi_address is None:
            ftdi_address = self.ftdi_address
        else:
            self.ftdi_address = ftdi_address
        if self.ftdi_address is not None:
            try:
                self.spi = SpiController(cs_count=4)
                self.spi.configure(ftdi_address + '/1')
                self.i2c = pyftdi.i2c.I2cController()
                self.i2c.configure(ftdi_address + '/2', frequency=5e4)
                self.pwm_controller.connect()
                self.valves.connect()
                self.stirrers.connect()
                self.photodiodes.connect()
                self.lasers.connect()
                self.thermometers.connect()
                self.pump1.connect()
                self.pump2.connect()
                self.pump3.connect()
                self.pump4.connect()
                print("Device %s connection established" % ftdi_address)
                self.dilution_worker = QueueWorker(device=self, worker_name="dilution")
                self.od_worker = QueueWorker(device=self, worker_name="od")
                if self.directory is not None and fit_calibration:
                    try:
                        self.fit_calibration_functions()
                        self.check_parameters()
                    except:
                        print("Calibration incomplete")
                    self.save()
            except pyftdi.ftdi.FtdiError:
                print("%s \nwith address %s is not connected!" % (str(self.__class__), ftdi_address))
            except pyftdi.usbtools.USBError:
                print("Device connected but does not recognize the command.\n   Please restart the Python session.")
            except pyftdi.usbtools.UsbToolsError:
                print("Device %s not connected." % ftdi_address)
        else:
            print("Please specify ftdi_address.")

    def pump_to(self, vial, p1=0, p2=0, p3=0, p4=0):
        self.valves.open(vial)  # valve number might be different for e.g. a lagoon setup
        if p1 > 0:
            self.pump1.pump(p1)
        if p2 > 0:
            self.pump2.pump(p2)
        if p3 > 0:
            self.pump3.pump(p3)
        if p4 > 0:
            self.pump4.pump(p4)
        while self.is_pumping():
            time.sleep(0.5)
        time.sleep(0.5)
        self.valves.close(vial)  # valve number might be different for e.g. a lagoon setup

    def update_cultures(self):
        def queued_function():
            for v, c in self.cultures.items():
                if self.soft_stop_trigger:
                    break
                if c is not None:
                    c.update()

        if self.dilution_worker.queue.empty():
            self.dilution_worker.queue.put(queued_function)
        else:
            print("Culture update not queued. dilution thread queue is not empty.")

    @property
    def pumps(self):
        return {1: self.pump1,
                2: self.pump2,
                3: self.pump3,
                4: self.pump4}

    def make_dilution(self, vial, pump1_volume=0, pump2_volume=0, pump3_volume=0, extra_vacuum=5):
        make_dilution(device=self, vial=vial, pump1_volume=pump1_volume, pump2_volume=pump2_volume,
                      pump3_volume=pump3_volume, extra_vacuum=extra_vacuum)

    def measure_od_all(self):
        vials_to_measure = []
        for v in self.cultures.keys():
            culture = self.cultures[v]
            if culture is not None:
                vials_to_measure += [v]

        def queued_function():
            measure_od_all(device=self, vials_to_measure=vials_to_measure)
        if self.od_worker.queue.empty():
            self.od_worker.queue.put(queued_function)
        else:
            print("OD measurement not queued. od thread queue is not empty.")

    def measure_temperature(self):
        def queued_function():
            self.thermometers.measure_temperature()
        if self.od_worker.queue.empty():
            self.od_worker.queue.put(queued_function)
        else:
            print("Temperature measurement not queued. od thread queue is not empty.")

    def prime_tubing(self, vial_number, pump_number, volume):
        self.lock_pumps.acquire()
        self.locks_vials[vial_number].acquire()
        self.valves.open(valve=vial_number)
        for i in range(20):
            q = input("%.2f ml of air available in pump %d tubing to vial %d?" % (volume,pump_number,vial_number))
            if q == "":
                self.pumps[pump_number].pump(volume=volume)
                while self.is_pumping():
                    time.sleep(0.5)
            else:
                break
        self.valves.close(valve=vial_number)
        self.lock_pumps.release()
        self.locks_vials[vial_number].release()

    def flush_tubing(self):
        """
        to prevent tube drying
        """
        def queued_function():
            for v in range(1, 8):
                c = self.cultures[v]
                if c is not None:
                    c.flush_tubing_if_necessary()
        self.dilution_worker.queue.put(queued_function)

    def is_pumping(self):
        return any(p.is_pumping() for p in [self.pump1, self.pump2, self.pump3, self.pump4])

    def emergency_stop(self):
        self.hard_stop_trigger = True
        self.pump1.stop()
        self.pump2.stop()
        self.pump3.stop()
        self.pump4.stop()
        self.pwm_controller.stop_all()

    def is_connected(self):
        """
        tries to send spi command. If device does not respond, returns False.
        """
        try:
            self.pump1.is_busy()
            return True
        except:
            return False

    def save(self):
        """
        saves calibration data and stock concentrations to self.directory/device_config.yaml
        """
        config_path = os.path.join(self.directory, "device_config.yaml")
        save_object(self, filepath=config_path)

    def load(self):
        self.file_lock.acquire()
        config_path = os.path.join(self.directory, "device_config.yaml")
        load_config(self, filepath=config_path)
        self.file_lock.release()
        print("Loaded device config.")
        for v in range(1, 8):
            vial_directory = os.path.join(self.directory, "vial_%d" % v)
            vial_config_path = os.path.join(vial_directory, "culture_config.yaml")
            self.cultures[v] = load_object(vial_config_path)
        print()

    def calibrate(self, dummy_data=False):
        if dummy_data:
            self.calibration_mv_to_od = {1: {1: 5, 10: 2, 22: 0.5, 35: 0.1, 41: 0.0001},
                                         2: {1: 5, 10: 2, 22: 0.5, 35: 0.1, 41: 0.0001},
                                         3: {1: 5, 10: 2, 35: 0.1, 41: 0.0001},
                                         4: {1: 5, 10: 2, 35: 0.1, 41: 0.0001},
                                         5: {1: 5, 10: 2, 35: 0.1, 41: 0.0001},
                                         6: {1: 5, 10: 2, 35: 0.1, 41: 0.0001},
                                         7: {1: 5, 10: 2, 35: 0.1, 41: 0.0001}}

            self.calibration_fan_speed_to_duty_cycle = {1: {1: 0.3, 2: 0.6, 3: 1},
                                                        2: {1: 0.3, 2: 0.6, 3: 1},
                                                        3: {1: 0.3, 2: 0.6, 3: 1},
                                                        4: {1: 0.3, 2: 0.6, 3: 1},
                                                        5: {1: 0.3, 2: 0.6, 3: 1},
                                                        6: {1: 0.3, 2: 0.6, 3: 1},
                                                        7: {1: 0.3, 2: 0.6, 3: 1}}
            self.calibration_pump_rotations_to_ml = {1: {1: 0.184, 3: 0.55, 20: 3.59, 80: 14.23},
                                                     2: {1: 0.184, 3: 0.55, 20: 3.59, 80: 14.23},
                                                     3: {},
                                                     4: {1: 0.17, 5: 0.64, 10: 1.05, 50: 5.1, 100: 7.55, 200: 13.71}}
            self.save()
            self.fit_calibration_functions()









    def show_parameters(self):
        self.stirrers.show_parameters()
        print()
        self.pump1.show_parameters()
        self.pump2.show_parameters()
        self.pump3.show_parameters()
        self.pump4.show_parameters()
        print()
        for od_sensor in self.od_sensors.values():
            od_sensor.calibration_curve()
            plt.show()

    def load_calibration(self, config_path):
        assert config_path.endswith("device_config.yaml")
        config = open(config_path).read()
        loaded_dict = yaml.load(config, Loader=yaml.Loader)
        # assert loaded_dict["_class"] == self.__class__
        for k in loaded_dict.keys():
            if k != "directory":
                self.__dict__[k] = loaded_dict[k]
        print("Loaded calibration data from %s" % config_path)
        self.fit_calibration_functions()

    def copy_calibration(self, config_path):
        self.load_calibration(config_path=config_path)
        self.check_parameters()
        self.save()

    def calibrate_pump(self, pump_number):
        calibrate_pump(device=self, pump_number=pump_number)

    def fill_vials(self, volume=15, vials=(1, 2, 3, 4, 5, 6, 7)):
        for v in vials:
            assert 1 <= v <= 7
        for v in vials:
            self.locks_vials[v].acquire()
            self.lock_pumps.acquire()
            self.valves.open(v)
            self.pump1.pump(volume)
            while self.is_pumping():
                time.sleep(0.5)
            time.sleep(2)
            assert not self.is_pumping()
            self.valves.close(v)
            self.locks_vials[v].release()
            self.lock_pumps.release()

    def aspirate_vials(self, volume=24, vials=(1, 2, 3, 4, 5, 6, 7)):
        for v in vials:
            assert 1 <= v <= 7
        for v in vials:
            self.locks_vials[v].acquire()
            self.lock_pumps.acquire()
            self.valves.open(v)
            self.pump4.pump(volume)
            while self.is_pumping():
                time.sleep(0.5)
            time.sleep(2)
            assert not self.is_pumping()
            self.valves.close(v)
            self.locks_vials[v].release()
            self.lock_pumps.release()

    def check_all(self):
        assert not self.hard_stop_trigger
        assert not self.soft_stop_trigger
        if not self.is_connected():
            self.connect()
        assert self.is_connected()
        assert not self.file_lock.locked()
        assert not self.lock_pumps.locked()
        for lock in self.locks_vials.values():
            assert not lock.locked()
        assert not self.is_pumping()
        self.check_parameters()
        self.check_cultures()

    def check_parameters(self):
        assert self.pump2.stock_concentration
        assert self.pump1.stock_volume
        assert self.pump2.stock_volume
        assert self.pump4.stock_volume
        for v in self.cultures.keys():
            if self.cultures[v] is not None:
                assert callable(self.od_sensors[v].calibration_function)
        assert callable(self.pump1.calibration_function)
        assert callable(self.pump2.calibration_function)
        # assert callable(self.device.pump3.calibration_function)
        assert callable(self.pump4.calibration_function)
        print("Calibration functions OK")

    def check_cultures(self):
        for v in self.cultures.keys():
            if self.cultures[v] is not None:
                self.cultures[v].check()

    def fit_calibration_functions(self):
        for v in range(1, 8):
            self.od_sensors[v].fit_calibration_function()
        self.pump1.fit_calibration_function()
        self.pump2.fit_calibration_function()
        self.pump3.fit_calibration_function()
        self.pump4.fit_calibration_function()



import logging
import os
import queue
import time
import threading


class ExperimentWorker:
    def __init__(self, experiment):
        self.experiment = experiment
        self.thread = None
        self.logger = logging.Logger(name="run_thread")
        log_filename = "main_run.log"
        log_filepath = os.path.join(experiment.directory, log_filename)
        handler = logging.FileHandler(log_filepath, 'a', 'utf-8')
        formatter = logging.Formatter('%(asctime)s.%(msecs)03d %(levelname)s %(module)s - %(funcName)s: %(message)s',
                                      "%Y-%m-%d %H:%M:%S")
        handler.setFormatter(formatter)
        self.logger.setLevel(logging.DEBUG)
        self.logger.addHandler(handler)
        self.keep_running = True

    def start(self):
        assert not self.is_alive()
        self.keep_running = True
        self.thread = threading.Thread(target=self._run_schedule, daemon=False)
        self.thread.start()

    def is_alive(self):
        if self.thread is None:
            return False
        else:
            return self.thread.is_alive()

    def status(self):
        if self.is_alive():
            print(time.ctime(), "Main thread: RUNNING")
        else:
            print(time.ctime(), "Main thread: NOT RUNNING")
        if self.experiment.device is not None:
            if self.experiment.device.od_worker is not None and self.experiment.device.dilution_worker is not None:
                self.experiment.device.od_worker.check_status()
                self.experiment.device.dilution_worker.check_status()
            else:
                print("Background thread workers not initialized.")
            for vial, lock in self.experiment.device.locks_vials.items():
                if lock.locked():
                    print("Vial %d: LOCKED" % vial)
            if self.experiment.device.lock_pumps.locked():
                print("Pumps: LOCKED")

    def stop(self):
        self.keep_running = False
        self.experiment.device.soft_stop_trigger = True
        while self.thread.is_alive():
            time.sleep(0.5)
        self.status()
        while (self.experiment.device.od_worker.is_performing_operation or
               self.experiment.device.dilution_worker.is_performing_operation):
            self.status()
            time.sleep(5)
        print(time.ctime(), "main thread and workers safely STOPPED")
        self.experiment.device.soft_stop_trigger = False

    def emergency_stop(self):
        self.experiment.device.emergency_stop()
        print("EMERGENCY STOP EXECUTED")
        self.stop()

    def _run_schedule(self):
        while self.keep_running:
            self.experiment.schedule.run_pending()
            time.sleep(1)


class QueueWorker:
    def __init__(self, device, worker_name):
        self.name = worker_name
        self.device = device
        self.logger = None
        self.queue = queue.Queue(maxsize=1)
        self.thread = threading.Thread(target=self.process_queue,
                                       args=[self.queue], daemon=False)
        if self.device.directory is not None:
            self.logger = logging.Logger(name=worker_name)
            log_filename = worker_name + '_thread.log'
            log_filepath = os.path.join(device.directory, log_filename)
            handler = logging.FileHandler(log_filepath, 'a', 'utf-8')
            formatter = logging.Formatter('%(asctime)s.%(msecs)03d %(levelname)s %(module)s - %(funcName)s: %(message)s',
                                          "%Y-%m-%d %H:%M:%S")
            handler.setFormatter(formatter)
            self.logger.setLevel(logging.DEBUG)
            self.logger.addHandler(handler)
            self.logger.info("Started Logger")
        self.is_performing_operation = False
        self.thread.start()

    def check_status(self):
        if not self.queue.empty():
            print("%s worker queue: NOT EMPTY" % self.name)
        if self.is_performing_operation:
            print(time.ctime(), "%s worker: WORKING" % self.name)
            # else:
            #     print("%s worker queue is empty." % self.name)
        else:
            print("%s worker: IDLE" % self.name)

    def process_queue(self, q):
        while True:
            try:
                queued_operation = q.get_nowait()
                try:
                    self.is_performing_operation = True
                    queued_operation()
                    self.is_performing_operation = False
                except Exception:
                    self.logger.exception("**** Fatal ERROR in loop ****")
                    self.is_performing_operation = False

            except queue.Empty:
                time.sleep(0.5)

    #
    #     self.keep_running = True
    #     self.start()
    #
    # def start(self):
    #     self.keep_running = True
    #     self.thread = threading.Thread(target=self.process_queue,
    #                                    args=[self.queue], daemon=False)
    #     self.thread.start()
    #
    # def is_alive(self):
    #     if self.thread is None:
    #         return False
    #     else:
    #         return self.thread.is_alive()
    #
    #
    # def stop(self):
    #     self.keep_running = False
    #     while self.thread.is_alive():
    #         time.sleep(0.5)
    #     print("%s worker stopped safely." % self.name)





from .turbidostat import TurbidostatCulture

from src.replifactory import increase_stress_drug1
import time


class PatientCulture(TurbidostatCulture):
    """
    class that simulates a patient taking antibiotics
        EXAMPLE:
    culture = EnduranceStressCulture(directory="./NewExperiment/", vial_number=1,
                                     name="Species 1", description="Strain 1", device=None,
                                     default_dilution_volume=10, dead_volume=15, od_max_limit=0.3,
                                     stress_increase_per_generation=1.05, initial_generations=3)
    """
    pumps = (1, 2, 4)

    def __init__(self, directory: str = None, vial_number: int = None,
                 name: str = "Patient", description: str = "Patient description",
                 default_dilution_volume: float = 10, dead_volume: float = 15):
        super().__init__(directory=directory, vial_number=vial_number, name=name, description=description,
                         default_dilution_volume=default_dilution_volume, dead_volume=dead_volume)
        self._treatment_started = False
        self._treatment_start_time = None
        self._treatment_stop_time = None
        self._current_drug_dose = 0
        self.od_threshold_treatment_start = 1.0
        self.od_threshold_dose_increase = 0
        self.dilution_period_hrs = 1.0
        self.treatment_dose = 40
        self.treatment_duration_hrs = 24
        self.treatment_dose_increase_delay_hrs = 0

        if hasattr(self, "od_max_limit"):
            del self.od_max_limit

        self.pumps = (1, 2, 4)
        # self.text = "Every 1 hours the 15 ml vial is diluted by adding 10ml and removing 10ml of clean medium." \
        #             "If OD>1, treatment begins, the drug concentration in the vial is set to 40mM and " \
        #             "maintained for 24h. After the treatment ends the dilutions are made with clean medium and " \
        #             "the drug concentration in the vial lowers with every dilution"

    def update(self):
        if self.is_active():
            if not bool(self._treatment_started):
                if self.od > self.od_threshold_treatment_start:  # trigger treatment start
                    self._treatment_started = True
                    self._current_drug_dose = self.treatment_dose
                    self._treatment_start_time = time.time()
                    self._treatment_stop_time = self._treatment_start_time + 3600 * self.treatment_duration_hrs
                    self.make_patient_dilution()

            else:
                if time.time() > self._treatment_stop_time:  # stop treatment, zero dose
                    self._current_drug_dose = 0
                if self.treatment_dose_increase_delay_hrs > 0:
                    if time.time()-self.treatment_dose_increase_delay_hrs*3600 > self._treatment_start_time:  # new dose
                        if self.od > self.od_threshold_dose_increase:
                            self._current_drug_dose *= 2
                            self._treatment_start_time = time.time()
                            self._treatment_stop_time = self._treatment_start_time + 3600 * self.treatment_duration_hrs
            if self.minutes_since_last_dilution > 60 * self.dilution_period_hrs:
                self.make_patient_dilution()

    def make_patient_dilution(self):
        increase_stress_drug1(culture=self, target_concentration=self._current_drug_dose)

    def check(self):
        super(TurbidostatCulture, self).check()
        assert self.treatment_dose <= self.device.pump2.stock_concentration


import time
import pandas as pd


class Stirrers:
    led_numbers = {stirrer: 7-stirrer for stirrer in [1, 2, 3, 4, 5, 6, 7]}

    def __init__(self, device):
        self.device = device
        self.pwm_controller = None
        self.multiplexer_port = None

        if self.device.is_connected():
            self.connect()

    def connect(self):
        self.pwm_controller = self.device.pwm_controller
        if self.pwm_controller.port is not None:
            self.multiplexer_port = self.device.i2c.get_port(self.device.PORT_GPIO_MULTIPLEXER_STIRRERS)  # for 3-pin fan speed feedback
            self.multiplexer_port.write_to(6, [0x00])  # all GPIO pins output
            self.multiplexer_port.write_to(7, [0x00])  # all GPIO pins output

    def show_parameters(self):
        df = pd.DataFrame.from_dict(self.device.calibration_fan_speed_to_duty_cycle)
        df.index = ["min duty cycle", "normal duty cycle", "max duty cycle"]
        df = df.T
        df.index.name = "Stirrer"
        for vial in range(1, 8):
            slow_speed = self.device.calibration_fan_speed_to_duty_cycle[vial][1]
            hi_speed = self.device.calibration_fan_speed_to_duty_cycle[vial][2]
            if slow_speed is None or hi_speed is None:
                string = "NOT CALIBRATED"
            else:
                too_slow = int(slow_speed * 100)
                working_range = int((hi_speed-slow_speed) * 100)
                too_fast = 100-too_slow-working_range

                string = "." * too_slow + "#" * working_range + "." * too_fast
                string = string[:60]
            df.loc[vial, "illustration"] = string
        print(df)

    def add_calibration_point(self, vial_number, speed, duty_cycle):
        assert speed in [0, 1, 2, 3]
        assert vial_number in [1, 2, 3, 4, 5, 6, 7]
        temp = self.device.calibration_fan_speed_to_duty_cycle
        temp[vial_number][speed] = duty_cycle
        self.device.calibration_fan_speed_to_duty_cycle = temp
        self.device.save()

    def set_rpm(self):
        pass  # TODO

    def _set_duty_cycle(self, vial, duty_cycle):
        assert 0 <= duty_cycle <= 1
        led_number = self.led_numbers[vial]
        self.pwm_controller.set_duty_cycle(led_number=led_number, duty_cycle=duty_cycle)

    def get_speed(self, vial):  # TODO
        self.pwm_controller.lock.acquire()
        duty_cycle = self.pwm_controller.get_duty_cycle(led_number=self.led_numbers[vial])
        self.pwm_controller.lock.release()
        if duty_cycle == 0:
            return 0
        elif duty_cycle == self.device.calibration_fan_speed_to_duty_cycle[vial][1]:
            return 1
        elif duty_cycle == self.device.calibration_fan_speed_to_duty_cycle[vial][2]:
            return 2
        else:
            return duty_cycle

    def set_speed(self, vial, speed, accelerate=True):
        assert speed in [0, 1, 2, 3]  # stopped, slow, fast
        self.check_calibration()
        if speed == 0:
            duty_cycle = 0
        else:
            duty_cycle = self.device.calibration_fan_speed_to_duty_cycle[vial][speed]

        self.pwm_controller.lock.acquire()
        try:
            if 0 < duty_cycle < 0.2 and accelerate:
                self._set_duty_cycle(vial, 1)
                time.sleep(0.1)

            self._set_duty_cycle(vial=vial, duty_cycle=duty_cycle)
        finally:
            self.pwm_controller.lock.release()

    def set_speed_all(self, speed, accelerate=True):
        assert speed in [0, 1, 2, 3]  # stopped, slow, normal, max
        self.check_calibration()
        self.pwm_controller.lock.acquire()
        try:
            for vial in range(1, 8):
                if speed == 0:
                    duty_cycle = 0
                else:
                    duty_cycle = self.device.calibration_fan_speed_to_duty_cycle[vial][speed]
                if duty_cycle < 0.2 and accelerate:
                    self._set_duty_cycle(vial=vial, duty_cycle=1)
                    time.sleep(0.1)
                self._set_duty_cycle(vial, duty_cycle=duty_cycle)
        finally:
            self.pwm_controller.lock.release()

    def check_calibration(self, vial=-1):
        if vial == -1:
            vials_list = [1, 2, 3, 4, 5, 6, 7]
        else:
            vials_list = [vial]
        for vial in vials_list:
            duty_cycle_slow = self.device.calibration_fan_speed_to_duty_cycle[vial][1]
            duty_cycle_fast = self.device.calibration_fan_speed_to_duty_cycle[vial][2]
            assert 0 <= duty_cycle_slow <= duty_cycle_fast <= 1

# def get_speed(vial_number=7,nbytes=4096,freq=1e5):
#     vial_number = 7 - vial_number
#     assert 0 <= vial_number <= 6
#     pm.slave_gpio_multiplexer.write_to(7, [0x00]) # 7 - port 1, 0x00 - output pin
#     pm.slave_gpio_multiplexer.write_to(2, [6 - vial_number])
#
#
#     fans.set_frequency(freq)
#     time.sleep(0.01)
#     res=fans.read(nbytes)
#
#     binstr="".join([bin(r)[2:].rjust(8,"0") for r in res])
#     binstr=binstr.replace("1110111", "1111111")
#     binstr=binstr.replace("11100111","11111111")
#
#     print(binstr,"\n")
#
#     sum(1 for i in binstr if i =="1")/(8*nbytes)
#
#     repeats = r'(.)\1+'
#
#     periods=[]
#     for match in re.finditer(repeats, binstr):
#         periods+=[len(match.group())]
#
#
#     periods=np.array(periods)
# #     print(periods)
#     if len(periods)>5:
#         periods=np.delete(periods,periods.argmin())
#         periods=np.delete(periods,periods.argmin())
#         periods=np.delete(periods,periods.argmin())
#     print("periods:",periods)
#
#     rpm=60*freq/periods.mean()/2
#     err=periods.std()/periods.mean()
#
# #     if err>0.06 or len(periods)<3:
# #         print('high error')
# #         time.sleep(3)
# #         if nbytes<=65280/5:
# #             nbytes*=2
# #         return get_speed(vial_number,nbytes,freq=freq)
#     print("Speed:",int(rpm),"RPM","±",int(err*rpm))
#     return rpm




import pandas as pd
from growth_rate import adaptive_window_doubling_time
from matplotlib import pyplot as plt


def plot_doubling_time_vs_od(time_values, od_values, dilution_timepoints=None):
    # global plt
    # if plt is None:
    #     import matplotlib.pyplot as plt

    # if od_values.min() <= 0:
    #     od_values = od_values-od_values.min()+1e-6
    od_values[od_values <= 0] = 1e-6

    df = pd.DataFrame(od_values, index=time_values, columns=["od"])

    ttd, td, err = adaptive_window_doubling_time(df.index.values, df.od.values, dilution_timepoints=dilution_timepoints)
    dftd = pd.DataFrame(td, index=ttd, columns=["td"])
    dftd.loc[:, "error"] = err
    dftd.index = dftd.index.astype(int)
    comb = df.merge(dftd, left_index=True, right_index=True)

    plt.figure(figsize=[12, 6])
    markers, caps, bars = plt.errorbar(comb.od, comb.td, yerr=comb.error, fmt="r.")
    [bar.set_alpha(0.5) for bar in bars]
    plt.ylim(0.3, 17)
    plt.xlim(-0.05, 1.5)
    plt.yscale("log")
    plt.yticks([0.5, 1, 2, 3, 4, 5, 10], labels=[0.5, 1, 2, 3, 4, 5, 10])
    plt.grid()
    plt.title("Effect of culture density on growth rate")
    plt.xlabel("Optical Density")
    plt.ylabel("Doubling time [hours]")